/* Copyright (C) 1991, 1994 Aladdin Enterprises.  All rights reserved.

   This file is part of Aladdin Ghostscript.

   Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
   or distributor accepts any responsibility for the consequences of using it,
   or for whether it serves any particular purpose or works at all, unless he
   or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
   License (the "License") for full details.

   Every copy of Aladdin Ghostscript must include a copy of the License,
   normally in a plain ASCII text file named PUBLIC.  The License grants you
   the right to copy, modify and redistribute Aladdin Ghostscript, but only
   under certain conditions described in the License.  Among other things, the
   License requires that the copyright notice and this notice be preserved on
   all copies.
 */

/* gdevrpdl.c */
/* RICOH RPDL driver */
#include "gdevprn.h"
#include "gp.h"		/* for gp_close_printer() */

#define	RPDL_MAX_BITMAP	32767
#define	RPDL_MAX_COPIES	99

#define rpdl_device(procs, dname, dpi, print_page_copies)\
{        std_device_std_color_full_body(gx_device_printer, &procs, dname,\
          (int)((long)(DEFAULT_WIDTH_10THS_A4) * (dpi) / 10),\
          (int)((long)(DEFAULT_HEIGHT_10THS_A4) * (dpi) / 10),\
          dpi, dpi,	/* X and Y resolution */\
	  1,		/* [bits/pixel] */\
          -(0) * (dpi),	/* Left offset */\
	  -(0) * (dpi),	/* Top offset */\
          (0) * 72.0,	/* Left margin */\
	  (0) * 72.0,	/* Bottom margin */\
          (0) * 72.0,	/* Right margin */\
	  (0) * 72.0	/* Top margin */\
        ),\
         { 0 },         /* std_procs */\
         { 0 },         /* skip */\
         { NULL, \
           print_page_copies,\
           gx_default_make_buffer_device,\
           gdev_prn_default_get_space_params,\
           gx_default_start_render_thread,\
           gx_default_open_render_device,\
           gx_default_close_render_device,\
           gx_default_buffer_page,\
           gdev_prn_get_overlay_bits,\
           gdev_prn_locate_overlay_buffer\
         },\
         { PRN_MAX_BITMAP, PRN_BUFFER_SPACE,\
             { band_params_initial_values },\
           0/*false*/,  /* params_are_read_only */\
           BandingAuto  /* banding_type */\
         },\
         { 0 },         /* fname */\
        0/*false*/,     /* OpenOutputFile */\
        0/*false*/,     /* ReopenPerPage */\
        0/*false*/, -1, /* Duplex[_set] */\
        0/*false*/, 0, 0, 0, /* file_is_new ... buf */\
        0, 0, 0, 0, 0/*false*/, 0, 0, /* buffer_memory ... clist_dis'_mask */\
        { 0 }      /* ... orig_procs */\
}

/* The device descriptors */
private dev_proc_put_params(rpdl_put_params);

private gx_device_procs
rpdl_procs = prn_params_procs(gdev_prn_open, gdev_prn_output_page,
	gdev_prn_close,gdev_prn_get_params,rpdl_put_params);

private dev_proc_print_page_copies(nx100_print_page_copies);

gx_device_printer far_data	/* fixed paper size */
gs_nx100f_device = rpdl_device(rpdl_procs, "nx100f", 600, nx100_print_page_copies);

gx_device_printer far_data	/* variable paper size */
gs_nx100v_device = rpdl_device(prn_std_procs, "nx100v", 600, nx100_print_page_copies);

/* ------ Internal routines ------ */

#define	SUB 0x1a
#define ESC 0x1b
#define	DC2 0x12
#define	EM  0x19

private const char
NX100Init[] = {
	ESC,0x0D, 'P',0x0D,				/* software reset */
/*	ESC, SUB, 'I',*/				/* printer reset */
	ESC, DC2, '!', '@', 'R', '0', '0', ESC, ' ',	/* RPDL mode */
	ESC, DC2, 'Y', 'W', ',', '3', ' ',		/* graphics (600dpi) */
	ESC, DC2, 'Y', 'A', '0', '4', ',', '3', ' ',	/* engine (600dpi) */
	ESC, DC2, 'Y', 'B', ',', '2', ' ',		/* max printable area */
	ESC, DC2, '#', '4', ' ',			/* coordination unit(600dpi) */
	ESC, DC2, 'Q', '5', ' ',			/* spaceing unit(600dpi) */
	ESC, DC2, 'M', '1', ' ',			/* normal CR,LF,FF */
	ESC, '!',					/* auto CR-LF disable */
	ESC, EM, 'M',					/* selecting multi feeder tray */
	ESC, DC2, '3', '3', ' ',			/* force exhausting time */
};

typedef struct rpdl_page_setup {
	uint		width, height;	/* [dots] */
	const char	*init;
	bool		landscape;	/* boolean */
} rpdl_page_setup;

private const rpdl_page_setup
RPDLPageSetup[] = {
	{(10500/254 + 2) * 60, (14850/254 + 2) * 60, "\033\022" "52@A6R\033 \033\022" "D1 ", 0},
	{(12850/254 + 2) * 60, (18200/254 + 2) * 60, "\033\022" "52@B6R\033 \033\022" "D1 ", 0},
	{(14850/254 + 2) * 60, (21000/254 + 2) * 60, "\033\022" "52@A5\033 \033\022" "D2 ", 1},
	{(55 + 2) * 60, (85 + 2) * 60, "\033\022" "52@HLR\033 \033\022" "D2 ", 1},
	{(18200/254 + 2) * 60, (25700/254 + 2) * 60, "\033\022" "52@B5\033 \033\022" "D2 ", 1},
	{(85 + 2) * 60, (110 + 2) * 60, "\033\022" "52@LT\033 \033\022" "D2 ", 1},
	{(21000/254 + 2) * 60, (29700/254 + 2) * 60, "\033\022" "52@A4\033 \033\022" "D2 ", 1},
	{(85 + 2) * 60, (140 + 2) * 60, "\033\022" "52@LGR\033 \033\022" "D1 ", 0},
	{(25700/254 + 2) * 60, (36400/254 + 2) * 60, "\033\022" "52@B4R\033 \033\022" "D1 ", 0},
	{(29700/254 + 2) * 60, (42000/254 + 2) * 60, "\033\022" "52@A3R\033 \033\022" "D1 ", 0},
	{(110 + 2) * 60, (170 + 2) * 60, "\033\022" "52@DLR\033 \033\022" "D1 ", 0},
	{max_uint, max_uint, "\033\022" "D1 ", 0}
};

/* run length compression for NX-100 */
private int
rpdl_compress(byte * rp, byte * cp, byte * end_data)
{
	int c = 0, count = 0;
	byte *saved_cp = cp;

	while (rp < end_data) {
		if ((*rp == c) && (count < 256)) {
			if (count == 0) {
				*cp++ = *rp;
			}
			count++;
		} else {
			if (count != 0) {
				*cp++ = count - 1;
				count = 0;
			}
			*cp++ = c = *rp;
		}
		rp++;
	}
	if (count)
		*cp++ = count - 1;

	return (cp - saved_cp);
}

/* Send the page to the printer.  */
private int
nx100_print_page_copies(gx_device_printer * pdev, FILE * pstream, int num_copies)
{
	const rpdl_page_setup	*setup_cmd = RPDLPageSetup;
	byte	*row_data, *comp_data;
	int	lnum, ncomp, nrow, line_size = gdev_prn_raster(pdev), nline = pdev->height;

	if (!(row_data = (byte *)gs_malloc(1, line_size*8, "LineBuffer(row)")))
		return_error(gs_error_VMerror);
	if (!(comp_data = (byte *)gs_malloc(1, line_size*12, "LineBuffer(compressed)")))
		return_error(gs_error_VMerror);

	fwrite(NX100Init, (sizeof NX100Init), 1, pstream);
	while ((pdev->width > setup_cmd->width) || (pdev->height > setup_cmd->height)) {
		setup_cmd++;
	}
	fputs(setup_cmd->init, pstream);
	fprintf(pstream, "\033\022" "N%i ", num_copies);
	if (setup_cmd->landscape) {
		nline &= ~7;
		for (lnum = 0; lnum < nline; lnum += 8) {
			register int i;
			byte	*in_ptr, *out_ptr, *bgn_data, *end_data;

			gdev_prn_copy_scan_lines(pdev, lnum, comp_data, line_size * 8);
			in_ptr = comp_data;
			out_ptr = row_data + line_size * 8 - 1;
			for (i = 0; i < line_size; i++) {
				gdev_prn_transpose_8x8(in_ptr, line_size, out_ptr, -1);
				out_ptr -= 8;
				in_ptr++;
			}
			bgn_data = row_data;
			end_data = row_data + line_size * 8;

			/* Remove leading 0s. */
			while (end_data > bgn_data && bgn_data[0] == 0)
				bgn_data++;

			/* Remove trailing 0s. */
			while (end_data > bgn_data && end_data[-1] == 0)
				end_data--;

			if ((nrow = end_data - bgn_data) == 0)
				continue;

			/* send the compressed data */
			ncomp = rpdl_compress(bgn_data, comp_data, end_data);
			if(ncomp >= nrow){
				fprintf(pstream, "\033\022" "G3,8,%d,,,%d,%d@",
				nrow, lnum, (bgn_data - row_data));
				fwrite(bgn_data, nrow, 1, pstream);
			}else{
				fprintf(pstream, "\033\022" "G3,8,%d,,4,%d,%d,%d@",
				nrow, lnum, (bgn_data - row_data), ncomp);
				fwrite(comp_data, ncomp, 1, pstream);
			}
		}
	} else {
		for (lnum = 0; lnum < nline; lnum++) {
			byte	*bgn_data, *end_data;

			gdev_prn_copy_scan_lines(pdev, lnum, row_data, line_size);
			bgn_data = row_data;
			end_data = row_data + line_size;

			/* Remove leading 0s. */
			while (end_data > bgn_data && bgn_data[0] == 0)
				bgn_data++;

			/* Remove trailing 0s. */
			while (end_data > bgn_data && end_data[-1] == 0)
				end_data--;

			if ((nrow = end_data - bgn_data) == 0)
				continue;

			/* send the compressed data */
			ncomp = rpdl_compress(bgn_data, comp_data, end_data);
			if(ncomp >= nrow){
				fprintf(pstream, "\033\022" "G3,%d,1,,,%d,%d@",
				nrow*8, (bgn_data - row_data) * 8, lnum);
				fwrite(bgn_data, nrow, 1, pstream);
			}else{
				fprintf(pstream, "\033\022" "G3,%d,1,,4,%d,%d,%d@",
				nrow*8, (bgn_data - row_data) * 8, lnum, ncomp);
				fwrite(comp_data, ncomp, 1, pstream);
			}
		}
	}
	fputc('\f', pstream);

	gs_free(row_data, 1, line_size*8, "LineBuffer(row)");
	gs_free(comp_data, 1, line_size*12, "LineBuffer(compressed)");

	return 0;
}

/* ================ Putting parameters ================ */
private int
rpdl_put_params(gx_device * pdev, gs_param_list * plist)
{
	int			ecode = 0, code, HWSize[2];
	gs_param_float_array	hwra, msa, psa;
	gs_param_int_array	hwsa;

	/* errors are checked later in gdev_prn_put_params(). */
	code = param_read_float_array(plist, "HWResolution", &hwra);
	if (code || (hwra.size != 2) || (hwra.data[0] <= 0) || (hwra.data[1] <= 0)){
		hwra.data = 0;	/* mark as not filled */
	}
	code = param_read_int_array(plist, "HWSize", &hwsa);
	if (code || (hwsa.size != 2) || (hwsa.data[0] <= 0) || (hwsa.data[1] <= 0)){
		hwsa.data = 0;	/* mark as not filled */
	}
	code = param_read_float_array(plist, ".MediaSize", &msa);
	if (code || (msa.size != 2) || (msa.data[0] <= 0) || (msa.data[1] <= 0)){
		msa.data = 0;	/* mark as not filled */
	}
	code = param_read_float_array(plist, "PageSize", &psa);
	if (code || (psa.size != 2) || (psa.data[0] <= 0) || (psa.data[1] <= 0)){
		psa.data = 0;	/* mark as not filled */
	}

	/*
	 * The HWResolution, HWSize, and MediaSize parameters interact in
	 * the following way:
	 *      1. Setting HWResolution recomputes HWSize from MediaSize.
	 *      2. Setting HWSize recomputes MediaSize from HWResolution.
	 *      3. Setting MediaSize recomputes HWResolution from HWSize.
	 * If more than one parameter is being set, we apply these rules
	 * in the order 1, 2, 3.  This does the right thing in the most
	 * common case of setting more than one parameter, namely,
	 * setting both HWResolution and HWSize.
	 */
	if (hwra.data != 0  &&
            (pdev->HWResolution[0] != hwra.data[0] || pdev->HWResolution[1] != hwra.data[1])) {
		pdev->HWResolution[0] = hwra.data[0];
		pdev->HWResolution[1] = hwra.data[1];
		pdev->width  = pdev->MediaSize[0] * hwra.data[0] / 72.0 + 0.5;
		pdev->height = pdev->MediaSize[1] * hwra.data[1] / 72.0 + 0.5;
	}
	if (hwsa.data != 0 &&
            (pdev->width != hwsa.data[0] || pdev->height != hwsa.data[1])) {
		pdev->width  = hwsa.data[0];
		pdev->height = hwsa.data[1];
		pdev->MediaSize[0] = hwsa.data[0] * 72.0 / pdev->HWResolution[0];
		pdev->MediaSize[1] = hwsa.data[1] * 72.0 / pdev->HWResolution[1];
	}
	/* .MediaSize takes precedence over PageSize */
	if (msa.data != 0 &&
            (pdev->MediaSize[0] != msa.data[0] || pdev->MediaSize[1] != msa.data[1])) {
		pdev->MediaSize[0] = msa.data[0];
		pdev->MediaSize[1] = msa.data[1];
		pdev->HWResolution[0] = pdev->width * 72.0 / msa.data[0];
		pdev->HWResolution[1] = pdev->height * 72.0 / msa.data[1];
	}else if (psa.data != 0 &&
            (pdev->MediaSize[0] != psa.data[0] || pdev->MediaSize[1] != psa.data[1])) {
		pdev->MediaSize[0] = psa.data[0];
		pdev->MediaSize[1] = psa.data[1];
		pdev->HWResolution[0] = pdev->width * 72.0 / psa.data[0];
		pdev->HWResolution[1] = pdev->height * 72.0 / psa.data[1];
	}
	/* commit */
	hwsa.persistent = psa.persistent = msa.persistent = hwra.persistent = false;
	if(hwra.data){
       		hwra.data = pdev->HWResolution;
		if((code = param_write_float_array(plist, "HWResolution", &hwra)) < 0)
			ecode = code;
	}
	if(hwsa.data){
       		HWSize[0] = pdev->width;
       		HWSize[1] = pdev->height;
		hwsa.data = HWSize;
		if((code = param_write_int_array(plist, "HWSize", &hwsa)) < 0)
			ecode = code;
	}
	if(msa.data){
       		msa.data = pdev->MediaSize;
		if((code = param_write_float_array(plist, ".MediaSize", &msa)) < 0)
			ecode = code;
	}
	if(psa.data){
       		psa.data = pdev->MediaSize;
		if((code = param_write_float_array(plist, "PageSize", &psa)) < 0)
			ecode = code;
	}
	if((code = gdev_prn_put_params(pdev, plist)) < 0){
		ecode = code;
	}
	/* we check the several parameters according to RPDLIV manual */
	if ((RPDL_MAX_BITMAP < pdev->width) || (RPDL_MAX_BITMAP < pdev->height)){
		ecode = gs_error_rangecheck;
	}
	if (pdev->NumCopies_set && (RPDL_MAX_COPIES < pdev->NumCopies)){
		ecode = gs_error_rangecheck;
	}
	return ecode;
}
